/* 
 * ufdbgclient.c
 *
 * light-weight process that connects to the ufdbguardd daemon to perform a URL verification.
 *
 * $Id: ufdbgclient.c,v 1.16 2006/12/07 10:09:51 root Exp root $
 */


#include "sg.h"

#include <stdio.h>
#include <time.h>
#include <libgen.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/tcp.h>

extern int    globalPid;
extern int    globalDebugRedirect;
extern char * globalLogDir;

int   noRedirects = 0;
char  serverName[256];
int   portNum;
int   be_quiet = 0;
int   httpsChecks = UFDB_API_HTTPS_CHECK_OFF;

void UFDBsetTunnelCheckMethod( int method )
{
   ;
}

/*
 * we use some code that is shared between the multithreaded ufdbguardd and
 * our single-threaded app.
 * Use pthread stubs to satisfy the linker without using the pthread overhead.
 */
int pthread_mutex_lock( void * dummy )
{
   return 0;
}

int pthread_mutex_trylock( void * dummy )
{
   return 0;
}

int pthread_mutex_unlock( void * dummy )
{
   return 0;
}

int UFDBcheckForHTTPStunnel( char * hostname, int portnumber, int flags )
{
   return UFDB_API_ERR_NULL;
}


int connect_to_server( void )
{
   int  s;
   struct hostent *    server;
   struct sockaddr_in  addr;
   struct linger       linger;

   errno = 0;

   /* 
    * check the server name.
    */
   server = gethostbyname( serverName );
   if (server == NULL)
   {
      char *  c; 			/* see if we have a dotted IP address */
      struct in_addr server_addr;

      c = serverName;
      while (*c != '\0')
      {
         if (*c != '.' && (*c < '0' || *c > '9'))
	 {
	    fprintf( stderr, "%s: hostname lookup failed for server '%s'\n", progname, serverName );
	    exit( 1 );
	 }
	 c++;
      }

      /* the name looks like a dotted IP address */
      server_addr.s_addr = inet_addr( serverName );
      server = gethostbyaddr( (const char *) &server_addr, sizeof(struct in_addr), AF_INET );
      if (server == NULL)
      {
         fprintf( stderr, "%s: IP lookup failed for server '%s'\n", progname, serverName );
	 exit( 1 );
      }
   }

   /*
    * create the socket to connect to the daemon.
    */
   s = socket( AF_INET, SOCK_STREAM, 0 );
   if (s < 0)
   {
      ufdbLogError( "cannot create socket: %s", strerror(errno) );
      exit( 1 );
   }

   addr.sin_family = AF_INET;
   memcpy( (void *) &addr.sin_addr, (void *) server->h_addr_list[0], sizeof(addr.sin_addr) );  
   addr.sin_port = htons( portNum );

   if (connect( s, (struct sockaddr *) &addr, sizeof(addr) ) < 0)
   {
      close( s );
      s = -1;
   }
   else
   {
      int            sock_parm;
      struct timeval tv;

      /*
       * Allow server-side addresses to be reused (don't have to wait for timeout).
       * Turn off NAGLE.
       */
#if 0
      /* this is only for the server: */
      sock_parm = 1;
      setsockopt( s, SOL_SOCKET, SO_REUSEADDR, (void *) &sock_parm, sizeof(sock_parm) );
#endif
      linger.l_onoff = 1;
      linger.l_linger = 2;
      setsockopt( s, SOL_SOCKET, SO_LINGER, (void *) &linger, sizeof(linger) );

      sock_parm = 1;
      setsockopt( s, IPPROTO_TCP, TCP_NODELAY, (void *) &sock_parm, sizeof(sock_parm) );
      sock_parm = 1;
      setsockopt( s, SOL_SOCKET, SO_KEEPALIVE, (void *) &sock_parm, sizeof(sock_parm) );

      /*
       * TCP buffers can be quite large.
       * Optimize the window size by setting it to a relatively small fixed size on the 
       * client and the server.
       */
      sock_parm = 16384;
      setsockopt( s, SOL_SOCKET, SO_SNDBUF, (void *) &sock_parm, sizeof(sock_parm) );
      sock_parm = 16384;
      setsockopt( s, SOL_SOCKET, SO_RCVBUF, (void *) &sock_parm, sizeof(sock_parm) );

      /*
       *  Prevent long blocking on communication with the daemon.
       */
      tv.tv_sec = 1 * 20;	/* 20 second timeout for reads */
      tv.tv_usec = 0;
      setsockopt( s, SOL_SOCKET, SO_RCVTIMEO, (void *) &tv, sizeof(tv) );

      tv.tv_sec = 1 * 20;	/* 20 second timout for writes */
      tv.tv_usec = 0;
      setsockopt( s, SOL_SOCKET, SO_SNDTIMEO, (void *) &tv, sizeof(tv) );
   }

   return s;
}


static void process_cmd( 
   int    fd, 
   char * url, 
   char * ip, 
   char * name )
{
   int    len;
   int    nbytes;
   char   line[MAX_BUF];

   if (strstr( url, "://" ) == NULL)
      sprintf( line, "http://%s %s/- %s GET\n", url, ip, name );
   else
      sprintf( line, "%s %s/- %s GET\n", url, ip, name );

   len = strlen( line );
   nbytes = write( fd, line, len );
   if (nbytes != len)
   {
      ufdbLogError( "line length is %d but wrote %d bytes.", len, nbytes );
      exit( 1 );
   }

   nbytes = read( fd, line, MAX_BUF );
   if (nbytes > MAX_BUF-1)
      nbytes = MAX_BUF - 1;
   if (nbytes < 0)
   {
      ufdbLogFatalError( "read from daemon failed: %s", strerror(errno) );
      fwrite( "\n", 1, 1, stdout );
   }
   else
   {
      line[nbytes] = '\0';            /* force null termination */
      fwrite( line, nbytes, 1, stdout );
   }
}


static void process_input( 
   int  server_socket  )
{
   int    nbytes;
   int    len;
   int    fd_stdout = fileno( stdout );
   int    num_waivers;
   time_t t0, t1;
   char   request[MAX_BUF];
   char   result[MAX_BUF];

   /*
    * loop to process queries on stdin.
    * If we cannot connect to the server, we continue looping:
    * - try to connect again to the server socket
    * - if failure, send an "OK" back to the requestor.
    */
   if (server_socket < 0)
   {
      ufdbLogError( "cannot connect to ufdbguardd daemon socket: %s", strerror(errno) );
   }
   
   num_waivers = 6;
   while (fgets(request, MAX_BUF, stdin) != NULL)	/* read request */
   {
      request[MAX_BUF-1] = '\0';
      len = strlen( request );

      if (UFDBglobalDebug)
	 ufdbLogError( "URL request: %s", request );

      if (server_socket < 0)
      {
	 num_waivers--;
	 if (num_waivers < 0)
	 {
	    num_waivers = 6;
	    server_socket = connect_to_server();		/* retry to connect after 6 lookups */
	    if (server_socket >= 0)
	    {
	       ufdbLogError( "communication with daemon re-established" );
	    }
	 }
      }

      if (server_socket < 0)		/* daemon is dead, say "OK" to caller */
      {
	 if (write( fd_stdout, "\n", 1 ) != 1)
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
	 continue;
      }

      errno = 0;
      nbytes = write( server_socket, request, len );		/* forward request */

      if (nbytes <= 0)  
      {								/* server socket was closed */
	 if (nbytes == 0)
	    errno = EINVAL;
         close( server_socket );
	 server_socket = -1;
	 num_waivers = 6;
	 ufdbLogError( "communication with ufdbguardd daemon lost (write): %s", strerror(errno) );
	 ufdbLogError( "all verifications will be rated \"OK\"" );
	 if (write( fd_stdout, "\n", 1 ) != 1)
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
	 continue;
      }

      if (nbytes != len)
      {
         close( server_socket );
	 server_socket = -1;
	 num_waivers = 6;
	 ufdbLogError( "line length is %d but wrote %d bytes to ufdbguardd (%s).", 
	               len, nbytes, strerror(errno) );
	 ufdbLogError( "all verifications will be rated \"OK\"" );
	 if (write( fd_stdout, "\n", 1 ) != 1)
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
	 continue;
      }

      errno = 0;
      t0 = time( NULL );
      nbytes = read( server_socket, result, MAX_BUF );		/* read result */
      t1 = time( NULL );
      if ((t1 - t0) > 1)
      {
         ufdbLogError( "server response is slow: nbytes=%d td=%d error=%s", 
	               nbytes, (t1 - t0), strerror(errno) );
      }

      if (nbytes <= 0)
      {								/* server socket was closed */
	 if (nbytes == 0)
	    errno = EINVAL;
         close( server_socket );
	 server_socket = -1;
	 num_waivers = 6;
	 ufdbLogError( "communication with ufdbguardd daemon lost (read): %s", strerror(errno) );
	 ufdbLogError( "all verifications will be rated \"OK\"" );
	 if (write( fd_stdout, "\n", 1 ) != 1)
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
	 continue;
      }

      if (nbytes > MAX_BUF - 1)
	 nbytes = MAX_BUF - 1;
      result[nbytes] = '\0';            	

      if (globalDebugRedirect)
      {
	 if (nbytes > 1)
	    ufdbLogError( "   REDIRECT %s %s GET", request, result );
      }

      if (noRedirects)					/* test mode */
      {
	 if (write( fd_stdout, "\n", 1 ) != 1)
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
      }
      else
      {
	 if (write( fd_stdout, result, nbytes ) != nbytes)	/* forward result */
	 {
	    ufdbLogError( "lost communication with parent (write): %s", strerror(errno) );
	    break;
	 }
      }

      if ((t1 - t0) >= 3)		/* !&*@^#$ why is the server slow ? */
      {
         close( server_socket );
	 server_socket = -1;
	 num_waivers = 0;
      }
   }

   if (UFDBglobalDebug)
      ufdbLogError( "end of input" );
}


static void client_usage( void )
{
   fprintf( stderr, "usage: ufdbgclient [-d] [-v] [-r] [-N] [-T] [-S server] [-p port] [-l logdir] [<URL> <IP> <user>]\n" );
   fprintf( stderr, "Options:\n" );
   fprintf( stderr, "  -d          : debug\n" );
   fprintf( stderr, "  -v          : show version number\n" );
   fprintf( stderr, "  -r          : log the redirections (performance impact)\n" );
   fprintf( stderr, "  -N          : No log file\n" );
   fprintf( stderr, "  -T          : test mode: log unwanted URLs but do not block them\n" );
   fprintf( stderr, "  -S server   : specify server name (default is %s)\n", "localhost" );
   fprintf( stderr, "  -p portnum  : override communication port (default is %d)\n", UFDB_DAEMON_PORT );
   fprintf( stderr, "  -l logdir   : log directory (default is %s)\n", DEFAULT_LOGDIR );
   fprintf( stderr, "  -q          : quiet mode\n" );
   fprintf( stderr, "  <URL>       : e.g. www.urlfilterdb.com\n" );
   fprintf( stderr, "  <IP>        : e.g. 10.1.1.1\n" );
   fprintf( stderr, "  <user>      : e.g. root or '-'\n" );
}


/* 
 * we do NOT want to link with the bz2 library so we have some stubs.
 */
void BZ2_bzBuffToBuffDecompress()
{ }


int main( 
   int                 argc, 
   char **             argv, 
   char **             envp )
{
   int                 s;
   struct stat         stbuf;
   struct tms 	       timer;
   struct timeval      start_time;

   globalPid = getpid();
#if 0
   sprintf( progname, "ufdbgclient-%02d", globalPid % 64 );
#else
   strcpy( progname, "ufdbgclient" );
#endif

   UFDBtimerInit( &timer );

   strcpy( serverName, "localhost" );
   portNum = UFDB_DAEMON_PORT;

   while ((s = getopt(argc, argv, "NPhdrtvql:p:S:")) != EOF)
   {
      switch (s) {
      case 'd':
	 UFDBglobalDebug = 1;
	 break;
      case 'r':
	 globalDebugRedirect = 1;
	 break;
      case 'l':
	 globalLogDir = optarg;
	 if (stat( globalLogDir, &stbuf ) != 0)
	 {
	    fprintf( stderr, "ufdbgclient: error in -l %s: %s\n", globalLogDir, strerror(errno) );
	    exit( 1 );
	 }
	 if (!S_ISDIR(stbuf.st_mode))
	 {
	    fprintf( stderr, "ufdbgclient: %s is not a directory\n", globalLogDir );
	    exit( 1 );
	 }
	 break;
      case 'N':
         ufdbGlobalSetLogging( 0 );
	 break;
      case 'p':
         portNum = atoi( optarg );
	 if (portNum < 1)
	 {
	    fprintf( stderr, "ufdbgclient: port number must be > 0\n" );
	    exit( 1 );
	 }
	 break;
      case 'q':
	 be_quiet = 1;
         break;
      case 'v':
	 fprintf( stderr, "ufdbgclient: %s\n", VERSION );
	 exit( 0 );
	 break;
      case 'S':
         strcpy( serverName, optarg );
	 break;
      case 'T':
	 fprintf( stderr, "-T option found.  Going into test mode.\n" );
	 noRedirects = 1;		/* Test mode means noRedirects */
	 break;
      case '?':
      case 'h':
      default:
         client_usage();
	 exit( 0 );
      }
   }

   sgSetGlobalErrorLogFile();
   signal( SIGPIPE, SIG_IGN );
   
   s = connect_to_server();

   /* 
    * process one query from the command line ?
    */
   if (optind < argc - 2)
   {
      if (s < 0)
      {
	 ufdbLogError( "cannot connect to port %d on server %s: %s", 
	                portNum, serverName, strerror(errno) );
	 exit( 1 );
      }
      process_cmd( s, argv[optind], argv[optind+1], argv[optind+2] );
      close( s );
      exit( 0 );
   }

   if (!be_quiet)
   {
      gettimeofday( &start_time, NULL );
      ufdbLogError( "ufdbgclient " VERSION " started" );
   }

   process_input( s );

   if (!be_quiet)
   {
      gettimeofday( &start_time, NULL );
      ufdbLogError( "ufdbgclient " VERSION " finished" );
   }

   close( s );

   exit( 0 );
}

