/*
 * ufdbchkport.c - URLfilterDB
 *
 * ufdbGuard is copyrighted (C) 2005,2006 by URLfilterDB with all rights reserved.
 * ufdbGuard is based on squidGuard.
 *
 * RCS $Id: ufdbchkport.c,v 1.4 2006/12/07 10:40:01 root Exp root $
 */

#include "sg.h"
#include "ufdb.h"
#include "ufdblib.h"
#include "ufdbchkport.h"
#include "httpsQueue.h"

#include "hashtable.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/tcp.h>

#if defined(__FreeBSD__)
#include <sys/select.h>
#endif

#ifdef UFDB_USE_SSL
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#if defined(UFDB_DO_DEBUG) || 0
#define DEBUG(x) fprintf x 
#else
#define DEBUG(x) 
#endif


int httpsChecks = UFDB_API_HTTPS_CHECK_OFF;

static pthread_mutex_t https_mutex = PTHREAD_MUTEX_INITIALIZER;
static SSL_CTX * ssl_ctx = NULL;
static struct hashtable * myht = NULL;

static int  lookupHTTPScache( char * hostname, int portnumber );
static int  UFDBverifyPortHasHTTPS( char * hostname, int portnumber );
static void addHTTPScache( char * hostname, int portnumber, int status );

static int ssl_connect( int fd, SSL ** ssl );
static int openssl_write( int fd, char * buf, int bufsize, void * ctx );
static int openssl_read( int fd, char * buf, int bufsize, void * ctx );
static void openssl_close( int fd, void * ctx );


int UFDBloadAllowedHTTPSsites( char * filename )
{
   return UFDB_API_OK;
}


static unsigned int hostname2hash( void * key )
{
   int n;
   unsigned int hash;
   char * hostname = (char *) key;

   hash = 0;
   n = 15;
   while (n > 0  &&  *hostname != '\0')
   {
      hash = (hash << 2) + ((unsigned int) (*hostname));
      hash ^= hash << 11;
      hostname++;
      n--;
   }

   return hash;
}


static int mystrcmp( void * a, void * b )
{
   return strcmp( a, b ) == 0;
}


void UFDBsetTunnelCheckMethod( int method )
{
   httpsChecks = method;
}


int UFDBinitHTTPSchecker( void )
{
   static int inited = 0;

   pthread_mutex_lock( &https_mutex );
   if (!inited)
   {
      inited = 1;
      myht = create_hashtable( 1024, hostname2hash, mystrcmp );
      srandom( ((getpid() << 8) ^ time(NULL)) + ((unsigned long)myht << 26) );
#ifdef UFDB_USE_SSL
      SSL_library_init();
      SSL_load_error_strings();
      ssl_ctx = SSL_CTX_new( SSLv23_client_method() );
      if (ssl_ctx == NULL)
      {
	 pthread_mutex_unlock( &https_mutex );
	 errno = EINVAL;
         return UFDB_API_ERR_ERRNO;
      }
      SSL_CTX_set_default_verify_paths( ssl_ctx );
      /*  SSL_CTX_load_verify_locations( ssl_ctx, opt.ca_cert, opt.ca_directory ); */

      /* SSL_VERIFY_NONE instructs OpenSSL not to abort SSL_connect if the
       * certificate is invalid.  We verify the certificate separately in
       * ssl_check_certificate, which provides much better diagnostics
       * than examining the error stack after a failed SSL_connect. 
       */
      SSL_CTX_set_verify( ssl_ctx, SSL_VERIFY_NONE, NULL );

      /* Since fd_write unconditionally assumes partial writes (and
       * handles them correctly), allow them in OpenSSL.  
       */
      SSL_CTX_set_mode( ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE );

      /* The OpenSSL library can handle renegotiations automatically, so
       * tell it to do so.
       */
      SSL_CTX_set_mode( ssl_ctx, SSL_MODE_AUTO_RETRY );
#endif
   }
   pthread_mutex_unlock( &https_mutex );

   return UFDB_API_OK;
}


static char * httpsGETroot( int s, char * hostname, int portnumber )
{
   SSL *       ssl;
   int         n;
   char        request[2048];
   char        reply[16384+2];

   /*******************************
   GET / HTTP/1.0
   User-Agent: Wget/1.8.2
   Host: www.urlfilterdb.com:9443
   Accept: * / *
   Connection: Keep-Alive
   ********************************/

   /* wget  --quiet -O - https://hostname:portnumber */

   if (ssl_ctx == NULL)
      return NULL;

   /* TODO: if there is a proxy, "CONNECT" must be used */

   n = ssl_connect( s, &ssl );
   if (n != UFDB_API_OK)
      return ufdbStrdup( "SSL connect failure" );

   sprintf( request, "GET / HTTP/1.0\r\n"
                     "User-Agent: Mozilla/4.0 "
		        "(compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.0.3705)\r\n"
		     "Host: %s:%d\r\n"
		     "Accept: */*\r\n"
		     "Connection: Close\r\n"
		     "\r\n",
		     hostname, portnumber );
   n = openssl_write( s, request, strlen(request), ssl );
   if (n < 0)
      strcpy( reply, "failed to send SSL message" );
   else
   {
      n = openssl_read( s, reply, sizeof(reply)-2, ssl );
      if (n > 0)
      {
	 reply[n] = '\0';
      }
      else
	 reply[0] = '\0';
   }
   openssl_close( s, ssl );

   return ufdbStrdup( reply );
}


/* Check for a tunnel.
 *
 * Valid flags are:
 * UFDB_API_ALLOW_QUEUING
 *
 * return values are:
 * UFDB_API_OK:          regular https traffic
 * UFDB_API_REQ_QUEUED:  request to queued for an other thread
 * UFDB_API_ERR_TUNNEL:  https channel is tunneled
 * UFDB_API_ERR_NULL:    tunnel verification is OFF.
 */
int UFDBcheckForHTTPStunnel( char * hostname, int portnumber, int flags )
{
   if (httpsChecks == UFDB_API_HTTPS_CHECK_OFF)
      return UFDB_API_ERR_NULL;

   UFDBinitHTTPSchecker();

   if (flags & UFDB_API_ALLOW_QUEUING)
   {
      int status;

      status = lookupHTTPScache( hostname, portnumber );
      if (status == UFDB_API_ERR_NULL)
      {
	 ufdbHttpsQueueRequest( hostname, portnumber );
	 return UFDB_API_REQ_QUEUED;
      }
      return status;
   }

   return UFDBverifyPortHasHTTPS( hostname, portnumber );
}


/*
 * in case that the caller of UFDBcheckForHTTPStunnel uses the UFDB_API_ALLOW_QUEUING flag,
 * the caller MUST have created 1-8 threads with UFDBhttpsTunnelVerifier() as main.
 */
void * UFDBhttpsTunnelVerifier( void * ptr )
{
   sigset_t  signals;
   int       portnumber;
   char      hostname[1028];

   /* 
    * The HUP signal must be blocked.
    */
   sigemptyset( &signals );
   sigaddset( &signals, SIGHUP );
   pthread_sigmask( SIG_BLOCK, &signals, NULL );

   ufdbSetThreadCPUaffinity();

   while (1)
   {
      ufdbGetHttpsRequest( hostname, &portnumber );
#if 0
      ufdbLogError( "UFDBhttpsTunnelVerifier: verifying %s:%d ...", hostname, portnumber );
#endif
      (void) UFDBverifyPortHasHTTPS( hostname, portnumber );
   }

   return NULL;  /* make the compiler happy */
}


/* Check that a port uses the https protocol.
 *
 * NOTE: This function may take 3 seconds to complete !
 */
static int UFDBverifyPortHasHTTPS( char * hostname, int portnumber )
{
   /* Hmmm. difficult task.
    * For now, we check that the port does not reply with 
    * - a SSH string indincating sshd and the use of proxytunnel
    */
   int            s;
   struct timeval tv;
   int            n;
   int            status;
   char *         content;
   char           line[132];

#if 0
   ufdbLogError( "UFDBverifyPortHasHTTPS( %s, %d )", hostname, portnumber );
#endif

   status = lookupHTTPScache( hostname, portnumber );
   if (status != UFDB_API_ERR_NULL)
   {
      return status;
   }

   s = UFDBopenSocket( hostname, portnumber );
   if (s < 0)
   {
      ufdbLogError( "https protocol verification for %s:%d failed: cannot open communication socket",
                     hostname, portnumber );
      addHTTPScache( hostname, portnumber, UFDB_API_ERR_SOCKET );
      return UFDB_API_ERR_SOCKET;
   }

   /* set the timeout on the socket: we wait for 3 seconds to find out if it sends "SSH-*" */
   tv.tv_sec = 3;
   tv.tv_usec = 0;
   setsockopt( s, SOL_SOCKET, SO_RCVTIMEO, (void *) &tv, sizeof(tv) );

   content = NULL;
   status = UFDB_API_OK;
   n = read( s, line, 8 );
   if (n > 0)
   {
      if (n >= 3  && strncmp( line, "SSH", 3 ) == 0)
      {
         ufdbLogError( "https protocol on %s:%d uses an SSH PROXY TUNNEL", hostname, portnumber );
	 status = UFDB_API_ERR_TUNNEL;
      }
      else if (n >= 3  && strncmp( line, "RFB", 3 ) == 0) 
      {
         ufdbLogError( "https protocol on %s:%d uses a VNC PROXY TUNNEL", hostname, portnumber );
	 status = UFDB_API_ERR_TUNNEL;
      }
      else
      {
	 int   i,j;
	 char  debuginfo[80];

	 /* very unlikely to be HTTPS since a norm HTTPS server start a handshake ! */
	 ufdbLogError( "https protocol on %s:%d does not encapsulate HTTP and is considered a PROXY TUNNEL",
	               hostname, portnumber );
	 if (n > 64)
	    n = 64;
	 for (j = 0, i = 0; i < n; i++)
	 {
	    if (line[i] == '\r'  ||  line[i] == '\n')
	       debuginfo[j++] = '_';
	    else if (isprint(line[i]))
	       debuginfo[j++] = line[i];
	    else
	       debuginfo[j++] = '.';
	 }
	 ufdbLogError( "   server started handshake with %s", debuginfo );
	 status = UFDB_API_ERR_TUNNEL;
      }
   }

   if (status == UFDB_API_OK)
   {
      /* TODO: check for vtunnel */

      /* setup a TLS connection to connect to the HTTPS port, do a "GET /", 
       * and see what the server has to say.
       */
      UFDBinitHTTPSchecker();


      /*******************************************************************
      ---request begin---
      GET / HTTP/1.0
      User-Agent: Wget/1.8.2
      Host: www.urlfilterdb.com:9443
      Accept: * / *
      Connection: Keep-Alive

      ---request end---
      HTTP/1.1 200 OK
      Date: Mon, 16 Oct 2006 16:18:04 GMT
      Server: BarracudaServer.com (Posix)
      Connection: Close
      Content-Type: text/html
      Cache-Control: No-Cache
      Set-Cookie: BarracudaDrive=3.0; expires=Sun, 12 Aug 2007 16:18:04 GMT
      ********************************************************************/

      content = httpsGETroot( s, hostname, portnumber );
      if (content == NULL  ||  *content == '\0')
      {
	 /* we did not read anything from the server...  so we cannot draw a conclusion.
	  * Therefore we return "OK" and hope that the next check gives and answer
	  * within the time limit (3 seconds default).
	  */
	 ufdbLogError( "https server %s:%d is lame", hostname, portnumber );
	 close( s );
	 ufdbFree( content );
	 return UFDB_API_OK;
      }

      if (strncmp( content, "HTTP/", 5 ) == 0)
      {
         if (UFDBglobalDebug)
	 {
	    int   i,j;
	    char  debuginfo[1080];

	    for (j = 0, i = 0; content[i] != '\0' && i < 1024; i++)
	    {
	       if (content[i] == '\r'  ||  content[i] == '\n')
		  debuginfo[j++] = '_';
	       else if (isprint(content[i]))
		  debuginfo[j++] = content[i];
	       else
		  debuginfo[j++] = '.';
	    }
	    debuginfo[j] = '\0';
	    ufdbLogError( "   https protocol reply: %s", debuginfo );
	 }

	 if (strstr( content, "Set-Cookie: SSLX_SSESHID=" ) != NULL)
	 {
	    ufdbLogError( "https protocol on %s:%d uses an SSL-EXPLORER TUNNEL", hostname, portnumber );
	    status = UFDB_API_ERR_TUNNEL;
	 }
	 else
	 if (strstr( content, "BarracudaServer.com" ) != NULL   ||
	     strstr( content, "barracudaserver.com" ) != NULL   ||
	     strstr( content, "BarracudaDrive" ) != NULL)
	 {
	    ufdbLogError( "https protocol on %s:%d uses a BARRACUDA PROXY TUNNEL", hostname, portnumber );
	    status = UFDB_API_ERR_TUNNEL;
	 }
	 else
	 if (strstr( content, "  index.vnc -" ) != NULL   ||
	     strstr( content, "  used with Xvnc" ) != NULL   ||
	     strstr( content, "TightVNC Java viewer applet" ) != NULL)
	 {
	    ufdbLogError( "https protocol on %s:%d uses a VNC PROXY TUNNEL", hostname, portnumber );
	    status = UFDB_API_ERR_TUNNEL;
	 }
      }
      else
      {
	 int   i,j;
	 char  debuginfo[80];

	 ufdbLogError( "https protocol on %s:%d encapsulates a non-HTTP protocol and "
		       "is considered a PROXY TUNNEL", 
		       hostname, portnumber );
	 for (j = 0, i = 0; content[i] != '\0' && i < 64; i++)
	 {
	    if (content[i] == '\r'  ||  content[i] == '\n')
	       debuginfo[j++] = '_';
	    else if (isprint(content[i]))
	       debuginfo[j++] = content[i];
	    else
	       debuginfo[j++] = '.';
	 }
	 debuginfo[j] = '\0';
	 ufdbLogError( "   https protocol reply: %s", debuginfo );

	 status = UFDB_API_ERR_TUNNEL;
      }
   }

   close( s );
   if (UFDBglobalDebug  ||  httpsChecks == UFDB_API_HTTPS_LOG_ONLY)
   {
      if (status == UFDB_API_OK)
	 ufdbLogError( "https protocol on %s:%d checked", hostname, portnumber );
      if (httpsChecks == UFDB_API_HTTPS_LOG_ONLY)
	 status = UFDB_API_OK;
   }
#if 0
   else
      ufdbLogError( "https protocol on %s:%d checked", hostname, portnumber );
#endif

   addHTTPScache( hostname, portnumber, status );

   ufdbFree( content );
   return status;
}


static int lookupHTTPScache( char * hostname, int portnumber )
{
   void * value;
   int    status;
   time_t now;
   char   key[1024+16];

   sprintf( key, "%s:%d", hostname, portnumber );

   pthread_mutex_lock( &https_mutex );
   value = hashtable_search( myht, (void *) key );
   if (value == NULL)
   {
      pthread_mutex_unlock( &https_mutex );
      return UFDB_API_ERR_NULL;
   }

   now = time( NULL );
   ((struct httpsInfo *) value)->t = now;
   status = ((struct httpsInfo *) value)->status;

#if 0
   ufdbLogError( "   lookupHTTPScache  %s:%d = %d", hostname, portnumber, status );
#endif

   if (status == UFDB_API_ERR_SOCKET  &&
       ((struct httpsInfo *) value)->t < now - 600)
   {
      value = hashtable_remove( myht, key );
      pthread_mutex_unlock( &https_mutex );
      ufdbFree( value );
      return UFDB_API_ERR_NULL;
   }

   pthread_mutex_unlock( &https_mutex );
   return status;
}


static void addHTTPScache( char * hostname, int portnumber, int status )
{
   unsigned int       n = 0;
   struct httpsInfo * info;
   char               key[1024+16];

#if 0
   ufdbLogError( "   addHTTPScache  %s:%d %d", hostname, portnumber, status );
#endif

   sprintf( key, "%s:%d", hostname, portnumber );
   info = ufdbMalloc( sizeof(struct httpsInfo) );
   info->t = time( NULL );
   info->status = status;

   n = hashtable_count( myht );
   pthread_mutex_lock( &https_mutex );
   if (n > 15000)			/* Hmmm. the hashtable is getting big ! */
   {
      hashtable_destroy( myht, 1 );
      myht = create_hashtable( 5000, hostname2hash, mystrcmp );
   }
   hashtable_insert( myht, (void *) ufdbStrdup(key), (void *) info );
   pthread_mutex_unlock( &https_mutex );

   /* UFDB_API_ERR_SOCKET: cache this for 20 minutes only */
}


int select_fd( int fd, double maxtime, int wait_for )
{
   int      result;
   fd_set   fdset;
   fd_set * rd;
   fd_set * wr;
   struct timeval tmout;

   FD_ZERO( &fdset );
   FD_SET( fd, &fdset );
   rd = &fdset;
   wr = &fdset;

   tmout.tv_sec = (long) maxtime;
   tmout.tv_usec = 1000000 * (maxtime - (long) maxtime);

   do
      result = select (fd + 1, rd, wr, NULL, &tmout);
   while (result < 0 && errno == EINTR);

   return result;
}


static void print_errors( void )
{
   unsigned long curerr = 0;
   int header = 0;

   while ((curerr = ERR_get_error()) != 0)
   {
      if (!header)
      {
	 ufdbLogError( "HTTPS proxy tunnel verification failed with OpenSSL error:" );
	 header = 1;
      }
      ufdbLogError( "   %s\n", ERR_error_string(curerr,NULL) );
   }
}


static int openssl_read( int fd, char * buf, int bufsize, void * ctx )
{
  int   ret;
  SSL * ssl = (SSL *) ctx;

  do
     ret = SSL_read( ssl, buf, bufsize );
  while (ret == -1  &&
	 SSL_get_error(ssl, ret) == SSL_ERROR_SYSCALL  &&
	 errno == EINTR);
  return ret;
}


static int openssl_write( int fd, char * buf, int bufsize, void * ctx )
{
  int   ret = 0;
  SSL * ssl = (SSL *) ctx;

  do
     ret = SSL_write( ssl, buf, bufsize );
  while (ret == -1  &&
	 SSL_get_error(ssl, ret) == SSL_ERROR_SYSCALL  &&
	 errno == EINTR);
  return ret;
}


#if 0
static int openssl_poll( int fd, double timeout, int wait_for, void * ctx )
{
  SSL * ssl = (SSL *) ctx;

  if (timeout == 0)
     return 1;
  if (SSL_pending(ssl))
     return 1;

  return select_fd( fd, timeout, wait_for );
}


static int openssl_peek( int fd, char * buf, int bufsize, void * ctx )
{
  int   ret;
  SSL * ssl = (SSL *) ctx;

  do
     ret = SSL_peek( ssl, buf, bufsize );
  while (ret == -1  &&
	 SSL_get_error(ssl, ret) == SSL_ERROR_SYSCALL  &&
	 errno == EINTR);
  return ret;
}
#endif


static void openssl_close( int fd, void * ctx )
{
  SSL * ssl = (SSL *) ctx;

  SSL_shutdown( ssl );
  SSL_free( ssl );

  close( fd );
}


/* Perform the SSL handshake on file descriptor FD, which is assumed
   to be connected to an SSL server.  The SSL handle provided by
   OpenSSL is registered with the file descriptor FD using
   fd_register_transport, so that subsequent calls to fd_read,
   fd_write, etc., will use the corresponding SSL functions.

   Returns UFDB_API_OK on success  */

static int ssl_connect( int fd, SSL ** ssl ) 
{
   if (ssl_ctx == NULL) 
      return UFDB_API_ERR_NULL;

   *ssl = SSL_new( ssl_ctx );
   if (*ssl == NULL)
      goto error;
   if (!SSL_set_fd(*ssl, fd))
      goto error;
   SSL_set_connect_state( *ssl );
   if (SSL_connect(*ssl) <= 0  ||  (*ssl)->state != SSL_ST_OK)
      goto error;

#if 0
   /* Register FD, i.e. arrange that our
      functions are used for reading, writing, and polling.  */
   fd_register_transport( fd, openssl_read, openssl_write, openssl_poll,
			  openssl_peek, openssl_close, ssl );
#endif

#if 0
   ufdbLogError( "connected socket %d to SSL handle 0x%08lx", fd, (unsigned long) *ssl );
#endif

   return UFDB_API_OK;

error:
#if 0
   ufdbLogError( "SSL handshake failed." );
#endif
   print_errors();
   if (*ssl != NULL)
      SSL_free( *ssl );
   return UFDB_API_ERR_NULL;
}

